#include "TState.h"
#include "TPoly.h"

class TTrellis
{
        int *metric;                                            // zaakumulowane metryki dla poszczeglnych stanw
        int **last_best;                                        // najlepszy poprzednik
        bool **transitions;                                     // jesli bylo przejscie ze stanu to ustawiane na true
        bool *reached_states;                                   // wskazuje, ktre stany zostay osignite
        int **distance;                                             // zakumulowany dystans dla poszczegolnych stanw

        TState *st;                                             // przechowuje stany systemu
        TPoly *G;                                               // przechowuje wielomiany
        TImage *img;

        int rzad;                                               // rzad systemu
        int states_count;                                       // liczba stanow
        int poly_count;                                         // liczba wyjsc
        int off;                                                // po ilu przejsciach zanikaja nieaktywne sciezki
        double snr;                                             // wspolczynik SNR

        // wspolrzedne
        int top;
        int left;
        int t_step;
        int l_step;

        AnsiString *input_seq;                                  // sekwencja wejciowa (zaszumiona)
        AnsiString *input_org;                                  // sekwencja oryginalna
        AnsiString in_original;
        int input_size;
        AnsiString dist;
        int diff_bits;                                          // ile bitw jest rnych w koderze i kanale


        void Init(void);
        void SetPolys(int poly_c, AnsiString *p);
        AnsiString GenOutStr(AnsiString in);
        void Step(AnsiString in, int t);
        int Szum(double we);

   public:

        void Create(int rank, int poly_c, AnsiString *p, TImage *im, double sn);
        void SetInputSeq(AnsiString in);
        void Calc(void);
        void Disp(int ile, int dis_off);
        void FindPath(int val);
        void DrawAll(int start);

        void SetOff(int o) { off = o; }
};

//------------------------------------------------------------------------------


void TTrellis::Init(void)
{
        reached_states = new bool[states_count];
        metric = new int[states_count];
        st = new TState[states_count];

        for(int i=0;i<states_count;i++)
        {
                reached_states[i] = false;
                metric[i] = 0;
                st[i].Init(i,rzad);
        }

}

void TTrellis::SetPolys(int poly_c, AnsiString *p)
{
        poly_count = poly_c;
        G = new TPoly[poly_count];

        for(int i=0;i<poly_count;i++)
        {
                G[i].SetRank(rzad);
                G[i].SetPoly(p[i]);
        }

}

AnsiString TTrellis::GenOutStr(AnsiString in)
{
        AnsiString out = "";
        for(int i=0;i<poly_count;i++)
                out = out + G[i].CalcOutput(in);

        return out;

}

void TTrellis::Create(int rank, int poly_c, AnsiString *p, TImage *im, double sn)
{
        rzad = rank;
        states_count = pow(2.0, rzad);
        img = im;
        snr = sn;
        top = 100;
        left = 80;
        //t_step = 320/states_count;

        Init();
        SetPolys(poly_c, p);

        t_step = 360/states_count;

        reached_states[0] = true;
        dist = "";

        //Step();
}

void TTrellis::SetInputSeq(AnsiString in)
{
        input_size = in.Length();
        in_original = in;
        input_seq = new AnsiString[input_size];
        input_org = new AnsiString[input_size];
        last_best = new int*[input_size];
        transitions = new bool*[input_size];
        distance = new int*[input_size];
        for(int i=0;i<input_size;i++)
        {
                last_best[i] = new int[states_count];
                transitions[i] = new bool[states_count];
                distance[i] = new int[states_count];
                for(int j=0;j<states_count;j++)
                {
                        last_best[i][j] = -1;
                        transitions[i][j] = false;
                        distance[i][j] = 0;
                }
        }

        int act_st = 0;
        diff_bits = 0;
        for(int i=1;i<=input_size;i++)
        {
                int c = st[act_st].Next(StrToInt(in.SubString(i,1)));
                AnsiString act_input = in.SubString(i,1) + st[act_st].GetState();
                AnsiString output_str = GenOutStr(act_input);
                input_org[i-1] = output_str;

                // tutaj doda, eby zaszumiao
                AnsiString noised_output="";
                for(int j=0;j<output_str.Length();j++)
                {
                        int act = StrToInt(output_str.SubString(j+1,1));
                        int w = Szum(StrToFloat(output_str.SubString(j+1,1)));
                        noised_output = noised_output+IntToStr(w);
                        if(w!=act) diff_bits++;
                }
                //input_seq[i-1] = output_str;
                input_seq[i-1] = noised_output;
                act_st = c;
        }

        if(input_size<=10) l_step = 600/input_size;
        else l_step=60;


//        CodeInput();
}

void TTrellis::Calc(void)
{
        for(int i=1;i<=input_size;i++)
        {
                Step(input_seq[i-1], i-1);
                /*
                for(int j=0;j<states_count;j++)
                {
                        if(st[j].GetMin()!=-1) img->Canvas->MoveTo(left+l_step*i, top+t_step*st[j].GetMin());
                        //img->Canvas->MoveTo(top+t_step*j, left+l_step*i);
                        img->Canvas->LineTo(left+l_step*(i-1), top+t_step*j);
                }
                */
        }
        DrawAll(0);
        /*
        for(int i=off;i<=input_size;i++)
        {
                Disp(i,off);
        }
        */


}

void TTrellis::DrawAll(int start)
{
        TRect rec(0,0,img->Width, img->Height);
        img->Canvas->Brush->Color = clCream;
        img->Canvas->FillRect(rec);

        img->Canvas->Font->Name = "Verdana";
        img->Canvas->Font->Style = TFontStyles()<< fsBold;
        img->Canvas->Font->Size = 10;
        img->Canvas->Font->Color = clNavy;

        AnsiString txt = "Wejcie :";
        int off_y = img->Canvas->TextHeight(txt);
        img->Canvas->TextOutA(10, -5+off_y, txt);

        txt = "Koder :";
        off_y = img->Canvas->TextHeight(txt);
        img->Canvas->TextOutA(10, 15+off_y, txt);

        txt = "Kana :";
        off_y = img->Canvas->TextHeight(txt);
        img->Canvas->TextOutA(10, 35+off_y, txt);

        /*
        txt = "Dek:";
        off_y = img->Canvas->TextHeight(txt);
        img->Canvas->TextOutA(10, 440+off_y, txt);
        */

        txt = "Liczba bitw przekamanych w kanale: ";
        off_y = img->Canvas->TextHeight(txt);
        img->Canvas->TextOutA(190, 460+off_y, txt);

        int proc = ((double)diff_bits/(poly_count*input_size))*100;
        txt = IntToStr(diff_bits) + "  (" + IntToStr(proc).SubString(1,3) + "% )";
        off_y = img->Canvas->TextHeight(txt);
        img->Canvas->TextOutA(500, 460+off_y, txt);


        int max = 10;
        int val;
        if(max<=input_size) val = max;
        else val = input_size;


        for(int t=start; t<start+val; t++)
        {


                int off_txt_y = -5;                             // przesunicie tekstu w pionie
                AnsiString txt = in_original.SubString(t+1,1);
                int off_x = img->Canvas->TextWidth(txt)/2;
                int off_y = img->Canvas->TextHeight(txt);
                img->Canvas->TextOutA(left+l_step*(t+1)-off_x-l_step/2, off_txt_y+off_y, txt);

                off_txt_y = 15;                             // przesunicie tekstu w pionie
                txt = input_org[t];
                off_x = img->Canvas->TextWidth(txt)/2;
                off_y = img->Canvas->TextHeight(txt);
                img->Canvas->TextOutA(left+l_step*(t+1)-off_x-l_step/2, off_txt_y+off_y, txt);

                off_txt_y = 35;                             // przesunicie tekstu w pionie
                txt = input_seq[t];
                off_x = img->Canvas->TextWidth(txt)/2;
                off_y = img->Canvas->TextHeight(txt);
                img->Canvas->TextOutA(left+l_step*(t+1)-off_x-l_step/2, off_txt_y+off_y, txt);
        }
}

void TTrellis::Disp(int ile, int dis_off)                               // ile to wskanik do punktu startu wywietlanej sekwencji
{
        TRect rec (70,80,img->Width, 440);
        img->Canvas->Brush->Color = clCream;
        img->Canvas->FillRect(rec);


        for(int t=ile-dis_off;t>0;t--)  // sprawdzanie wszystkich stanow od ile-offset i usuwanie nieskonczonych sciezek
        {
                bool *trans = new bool[states_count];
                for(int i=0;i<states_count;i++)
                        trans[i] = false;

                for(int i=0;i<states_count;i++) // dla wszystkich stanow w danej chwili czasu sprawdzenie czy nastapia tranzycja
                {
                        int c = last_best[t][i];
                        trans[c] = true;
                }

                for(int i=0;i<states_count;i++)
                {
                        if(!trans[i])
                                last_best[t-1][i] = -1;
                }
        }

        for(int t=0;t<ile;t++)
        {
        //        Step(input_seq[t], t);
                for(int j=0;j<states_count;j++)
                {
                        // last_best[t][j]
                        if(last_best[t][j]!=-1)
                        {
                                img->Canvas->Pen->Width = 1;
                                // rysowanie od "konca" czyli od stanu ostatniego
                                if(j<states_count/2)
                                {
                                        img->Canvas->Pen->Color = clBlue;
                                        img->Canvas->Pen->Style = psSolid;
                                }
                                else
                                {
                                        img->Canvas->Pen->Color = clRed;
                                        img->Canvas->Pen->Style = psDot;

                                }
                                // rysowanie kreski
                                img->Canvas->MoveTo(left+l_step*(t+1), top+t_step*j);
                                img->Canvas->LineTo(left+l_step*(t), top+t_step*last_best[t][j]);
                                // jaki tekst do wyswietlenia nad punktem kraty
                                AnsiString txt = "(" + IntToStr(distance[t][j]) + ")";
                                // wspolrzedne dla przesuniecia tekstu, zeby byl na srodku
                                int off_x = img->Canvas->TextWidth(txt)/2;
                                int off_y = img->Canvas->TextHeight(txt)+3;
                                // ustawienie parametrow czcionki
                                img->Canvas->Font->Name = "Verdana";
                                img->Canvas->Font->Style = TFontStyles()<< fsBold;
                                img->Canvas->Font->Size = 10;
                                img->Canvas->Font->Color = clGreen;
                                // wyswietlenie tekstu
                                img->Canvas->TextOutA(left+l_step*(t+1)-off_x, top+t_step*j-off_y, txt);
                        }
                }
        }

        int max = 10;
        int val;
        if(max<=input_size) val = input_size;
        else val = max;

        for(int t=0;t<=val;t++)
        {
                for(int j=0;j<states_count;j++)
                {
                        img->Canvas->Pen->Width = 5;
                        img->Canvas->Pen->Color = clBlack;
                        img->Canvas->MoveTo(left+l_step*t, top+t_step*j);
                        img->Canvas->LineTo(left+l_step*t, top+t_step*j);
                }
        }
        if(ile==10) FindPath(ile-1);
        return;
}

void TTrellis::FindPath(int val)
{
        int best_st=0;
        for(int i=1;i<states_count;i++)
        {
                if(distance[val][i]<distance[val][best_st])
                        best_st = i;
        }

        AnsiString decode = "";
        for(int t=val;t>=0;t--)
        {
                int p = last_best[t][best_st];  // pamita stan nastpny w odtwarzaniu do tylu
                img->Canvas->Pen->Width = 2;
                img->Canvas->MoveTo(left+l_step*(t+1), top+t_step*best_st);
                img->Canvas->LineTo(left+l_step*(t), top+t_step*last_best[t][best_st]);

                AnsiString txt;
                // wspolrzedne dla przesuniecia tekstu, zeby byl na srodku
                if(best_st<(states_count/2))
                {
                        txt = "0";
                        decode = "0"+decode;
                }
                else
                {
                        txt = "1";
                        decode = "1"+decode;
                }


                int off_x = img->Canvas->TextWidth(txt)/2;
                int off_y = img->Canvas->TextHeight(txt);
                img->Canvas->Font->Name = "Verdana";
                img->Canvas->Font->Style = TFontStyles()<< fsBold;
                img->Canvas->Font->Size = 10;
                img->Canvas->Font->Color = clRed;
                // wyswietlenie tekstu
                img->Canvas->TextOutA(left+l_step*(t+0.5)-off_x, 435+off_y, txt);
                // po zakonczeniu
                best_st = p;
        }

        AnsiString txt = "Dek:";
        int off_y = img->Canvas->TextHeight(txt);
        img->Canvas->TextOutA(10, 435+off_y, txt);

        int diff=0;
        for(int i=0;i<decode.Length();i++)
        {
                int d = StrToInt(decode.SubString(i+1,1));
                int org = StrToInt(in_original.SubString(i+1,1));
                if(d!=org) diff++;
        }
        img->Canvas->Font->Color = clNavy;
        txt = "Liczba bdw przy odbiorze:";
        off_y = img->Canvas->TextHeight(txt);
        img->Canvas->TextOutA(190, 480+off_y, txt);

        int proc = ((double)diff/input_size)*100;
        txt = IntToStr(diff) + "  (" + IntToStr(proc) + "% )";
        off_y = img->Canvas->TextHeight(txt);
        img->Canvas->TextOutA(500, 480+off_y, txt);
}

void TTrellis::Step(AnsiString in, int t)
{
        bool *states_update = new bool[states_count];
        for(int i=0;i<states_count;i++)
                states_update[i] = reached_states[i];

        for(int i=0;i<states_count;i++)
        {
                if(reached_states[i])
                {
                        for(int j=0;j<2;j++)
                        {
                                int c = st[i].Next(j);
                                //reached_states[c] = true;                               // zmienic, zeby update sie dokonywal po petli
                                states_update[c] = true;
                                AnsiString act_input = IntToStr(j) + st[i].GetState();
                                AnsiString output_str = GenOutStr(act_input);
                                int cost = st[i].TransCost(in, output_str);          // zmienic, zeby porownywal string z wejscia
                                st[c].SetMetric(i,st[i].CalcDist(cost));                // jako koszt przejscie ze stanu i zapisuje calkowity dotychczasowy koszt dla tego stanu plus koszt przejscia
                                //st[c].SetDist(acc_dist+cost);
                        }
                        // tutaj dodac zmienna ktora bedzie sygnalizowala dla calej klasy czy dotarlismy do wszystkich stanow
                }
        }

        for(int i=0;i<states_count;i++)
        {
                reached_states[i] = states_update[i];
                st[i].ComputeDist();
                //if(t>0)
                {
                        int c = st[i].GetMin();
                        last_best[t][i] = c;
                        transitions[t][c] = true;
                        distance[t][i] = st[i].GetAccDist();
                }
                // tutaj instrukcja zapamitania najlepszego poprzednika
                st[i].ClearMetric();
                dist += " " + IntToStr(st[i].GetAccDist());
        }
        dist += "\n";
        return;
}

int TTrellis::Szum(double we)
{
        double wy;
        we==1 ? wy=1.0 : wy=-1.0;

        double x = RAND_MAX;
        double ll = 0.0;
        for(int i=0;i<12;i++)
                ll += (double)random(x)/(x-1);

        double norm = ll-6;

        //double sig = pow(10, -snr/20);
        double sig = exp((-snr*log(10))/20);

        wy = wy + 0.5*norm*sig;
        wy<0 ? wy=0 : wy=1;
        return wy;
}
